"""This example script shows how to develop a simple bot with davtelepot.

1. Install davtelepot
    ```bash
    # Create a new python virtual environment
    pip -m venv env
    # Activate the virtual environment
    source env/bin/activate
    # Install davtelepot library in this environment
    pip install davtelepot
    # Run the current script within this environment
    python a_simple_bot.py
    ```

2. To run your bot, you will need a bot token. You can get up to 20 bot tokens
    from https://t.me/botfather

3. This script will look for your bot token in a gitignored `secrets.py` file
    in the same folder as the current script. If no `secrets` module is found,
    user will be prompted at runtime for a token, and the entry will be stored
    in `secrets.py` for later use.

4. Standard library, third party and project modules will be imported, and
    `logging` preferences set. You are free to edit these settings, e.g. adding
    one or more file and error loggers or setting different levels.

5. `simple_bot` is an instance of `davtelepot.bot.Bot` class.
    To instantiate a bot you need to provide at least a Telegram bot API token,
    and you may also provide a path to a database (if no path is provided,
    a `./bot.db` SQLite database will be used).

6. `initialize_bot` function is defined and called on `simple_bot`. It assigns
    commands, parsers, callback and inline query handlers to the bot.

7. `Bot.run()` method is called, causing the script to hang while asynchronous
    tasks run checking for updates and routing them to get and send replies.
    Send a KeyboardInterrupt (ctrl+C) to stop the bot.
"""

# Standard library modules
import logging
import os
import sys

# Third party modules
try:
    from davtelepot.bot import Bot
    from davtelepot.utilities import (
        get_cleaned_text, get_user, make_inline_keyboard, make_button
    )
except ImportError:
    logging.error(
        "Please install davtelepot library.\n"
        "Using a python virtual environment is advised.\n\n"
        "```bash\n"
        "pip -m venv env\n"
        "env/bin/pip install davtelepot\n"
        "env/bin/python davtelepot/examples/a_simple_bot.py"
        "```"
    )
    sys.exit(1)

# Get path of current script
path = os.path.dirname(os.path.abspath(__file__))


def initialize_bot(telegram_bot):
    """Take a bot and set commands."""
    telegram_bot.set_callback_data_separator('|')

    @telegram_bot.command(command='foo', aliases=['Foo'],
                          show_in_keyboard=True,
                          description="Reply 'bar' to 'foo'",
                          authorization_level='everybody')
    async def foo_command(bot, update, user_record):
        return (
            "Bar!\n\n"
            f"You wrote: {update['text']}\n"
            f"I am @{bot.name} and you are {get_user(user_record)}"
        )

    def is_bar_text_message(lowered_text):
        return lowered_text.startswith('bar')

    @telegram_bot.parser(condition=is_bar_text_message,
                         description='Reply Foo to users who write Bar',
                         authorization_level='everybody')
    async def bar_parser(bot, update):
        text_except_foo = get_cleaned_text(update, bot, ['bar'])
        return f"Foo!\n{text_except_foo}"

    def get_keyboard(chosen_button=-1):
        return make_inline_keyboard(
            [
                make_button(
                    prefix='button:///',
                    delimiter='|',
                    data=[i],
                    text=f"{'✅' if chosen_button == i else '☑️'} Button #{i}"
                )
                for i in range(1, 13)
            ],
            3
        )

    @telegram_bot.command(command='buttons')
    async def buttons_command():
        return dict(
            text="Press a button!",
            reply_markup=get_keyboard()
        )

    @telegram_bot.button(prefix='button:///', separator='|',
                         authorization_level='everybody')
    async def buttons_button(data):
        button_number = data[0]
        return dict(
            edit=dict(
                text=f"You pressed button #{button_number}",
                reply_markup=get_keyboard(button_number)
            )
        )

    def starts_with_a(text):
        return text.startswith('a')

    @telegram_bot.query(
        condition=starts_with_a,
        description='Mirror query text if it starts with letter `a`',
        authorization_level='everybody'
    )
    async def inline_query(update):
        return dict(
            type='article',
            id=10,
            title="Click here to send your query text as a message.",
            input_message_content=dict(
                message_text=update['query']
            )
        )

    telegram_bot.set_default_inline_query_answer(
        dict(
            type='article',
            id=0,
            title="Start query text with `a` to mirror it.",
            input_message_content=dict(
                message_text="This query does not start with `a`."
            )
        )
    )

    telegram_bot.set_unknown_command_message(
        "<b>Currently supported features</b>\n\n"
        "- /foo (or text starting with `foo`): replies `Bar!`.\n"
        "- Text starting with `bar`: replies `Foo!` followed by the rest of "
        "text in your bar-starting message.\n"
        "- /buttons demonstrates the use of buttons.\n"
        "- Inline queries starting with letter `a` will be mirrored as text "
        "messages. To use this feature, try writing <code>@{bot.name} "
        "your_text_here</code>"
    )


def _main():
    # Import or prompt user for bot token
    try:
        from secrets import simple_bot_token
    except ImportError:
        simple_bot_token = input("Enter bot token:\t\t")
        with open(
            f'{path}/secrets.py',
            'a'  # Append to file, create it if it does not exist
        ) as secrets_file:
            secrets_file.write(f'simple_bot_token = "{simple_bot_token}"\n')

    # Set logging preferences
    log_formatter = logging.Formatter(
        "%(asctime)s [%(module)-15s %(levelname)-8s]     %(message)s",
        style='%'
    )
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_formatter)
    console_handler.setLevel(logging.DEBUG)
    root_logger.addHandler(console_handler)

    # Instantiate, initialize and make `simple_bot` run.
    simple_bot = Bot(token=simple_bot_token, database_url=f"{path}/bot.db")
    initialize_bot(simple_bot)
    logging.info("Send a KeyboardInterrupt (ctrl+C) to stop bots.")
    Bot.run()


if __name__ == '__main__':
    _main()



"""Get and parse Telegram API web page."""

# Standard library modules
import argparse
import asyncio
import inspect
import logging

# Third party modules
import os
from typing import List

import aiohttp
from bs4 import BeautifulSoup

# Project modules
from davtelepot.api import TelegramBot

api_url = "https://core.telegram.org/bots/api"


class TelegramApiMethod(object):
    types = {
        'Array of String': "List[str]",
        'Boolean': "bool",
        'Integer': "int",
        'Integer or String': "Union[int, str]",
        'String': "str",
    }
    """Telegram bot API method."""

    def __init__(self, name, description, table):
        """Initialize object with name, description and table data."""
        self._name = name
        self._description = description
        self._table = table
        self._parameters = self.get_parameters_from_table()

    @property
    def name(self):
        """Return method name."""
        return self._name

    @property
    def description(self):
        """Return method description."""
        return self._description

    @property
    def table(self):
        """Return method parameters table."""
        return self._table

    @property
    def parameters(self):
        return self._parameters

    @property
    def parameters_with_types(self) -> List[str]:
        return [
            f"{parameter['name']}: {parameter['type']}"
            for parameter in self._parameters
        ]

    def print_parameters_table(self):
        """Extract parameters from API table."""
        result = ''
        if self.table is None:
            return "No parameters"
        rows = self.table.tbody.find_all('tr')
        if rows is None:
            rows = []
        for row in rows:
            result += '------\n'
            columns = row.find_all('td')
            if columns is None:
                columns = []
            for column in columns:
                result += f'| {column.text.strip()} |'
            result += '\n'
        result += '\n'
        return result

    def get_parameters_from_table(self):
        if self.table is None:
            return []
        parameters = []
        rows = self.table.tbody.find_all('tr') or []
        for row in rows:
            columns = row.find_all('td') or []
            name, type_, *_ = map(lambda column: column.text.strip(), columns)
            if type_ in self.types:
                type_ = self.types[type_]
            else:
                type_ = f"'{type_}'"
            parameters.append(
                dict(
                    name=name,
                    type=type_
                )
            )
        return parameters


async def print_api_methods(filename=None,
                            print_all=False,
                            output_file=None,
                            input_file=None):
    """Get information from Telegram bot API web page."""
    implemented_methods = dir(TelegramBot)
    if input_file is None or not os.path.isfile(input_file):
        async with aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(
                total=100
            )
        ) as session:
            async with session.get(
                    api_url
            ) as response:
                web_page = BeautifulSoup(
                    await response.text(),
                    "html.parser"
                )
    else:
        with open(input_file, 'r') as local_web_page:
            web_page = BeautifulSoup(
                ''.join(local_web_page.readlines()),
                "html.parser"
            )
    if filename is not None:
        with open(filename, 'w') as _file:
            _file.write(web_page.decode())
    methods = []
    for method in web_page.find_all("h4"):
        method_name = method.text
        description = ''
        parameters_table = None
        for tag in method.next_siblings:
            if tag.name is None:
                continue
            if tag.name == 'h4':
                break  # Stop searching in siblings if <h4> is found
            if tag.name == 'table':
                parameters_table = tag
                break  # Stop searching in siblings if <table> is found
            description += tag.get_text()
        # Methods start with a lowercase letter
        if method_name and method_name[0] == method_name[0].lower():
            methods.append(
                TelegramApiMethod(
                    method_name,
                    description,
                    parameters_table
                )
            )
    new_line = '\n'
    new_methods = []
    edited_methods = []
    for method in methods:
        if print_all or method.name not in implemented_methods:
            new_methods.append(method)
        else:
            parameters = set(parameter['name'] for parameter in method.parameters)
            implemented_parameters = set(
                parameter.strip('_')  # Parameter `type` becomes `type_` in python
                for parameter in inspect.signature(
                    getattr(TelegramBot,
                            method.name)
                ).parameters.keys()
                if parameter != 'self'
            )
            new_parameters = parameters - implemented_parameters
            deprecated_parameters = implemented_parameters - parameters - {'kwargs'}
            if new_parameters or deprecated_parameters:
                edited_methods.append(
                    dict(
                        name=method.name,
                        new_parameters=new_parameters,
                        deprecated_parameters=deprecated_parameters
                    )
                )
    if output_file:
        with open(output_file, 'w') as file:
            if new_methods:
                file.write(
                    "from typing import List, Union\n"
                    "from davtelepot.api import TelegramBot\n\n\n"
                    "# noinspection PyPep8Naming\n"
                    "class Bot(TelegramBot):\n\n"
                )
            file.writelines(
                f"    async def {method.name}("
                f"{', '.join(['self'] + method.parameters_with_types)}"
                f"):\n"
                f"        \"\"\""
                f"{method.description.replace(new_line, new_line + ' ' * 4)}\n"
                f"        See https://core.telegram.org/bots/api#"
                f"{method.name.lower()} for details.\n"
                f"        \"\"\"\n"
                f"        return await self.api_request(\n"
                f"            '{method.name}',\n"
                f"            parameters=locals()\n"
                f"        )\n\n"
                for method in new_methods
            )
            if edited_methods:
                file.write('\n# === EDITED METHODS ===\n')
            for method in edited_methods:
                file.write(f'\n"""{method["name"]}\n')
                if method['new_parameters']:
                    file.write("    New parameters: "
                               + ", ".join(method['new_parameters'])
                               + "\n")
                if method['deprecated_parameters']:
                    file.write("    Deprecated parameters: "
                               + ", ".join(method['deprecated_parameters'])
                               + "\n")
                file.write('"""\n')
    else:
        print(
            '\n'.join(
                f"NAME\n\t{method.name}\n"
                f"PARAMETERS\n\t{', '.join(['self'] + method.parameters_with_types)}\n"
                f"DESCRIPTION\n\t{method.description}\n"
                f"TABLE\n\t{method.print_parameters_table()}\n\n"
                for method in new_methods
            )
        )
        for method in edited_methods:
            print(method['name'])
            if method['new_parameters']:
                print("\tNew parameters: " + ", ".join(method['new_parameters']))
            if method['deprecated_parameters']:
                print("\tDeprecated parameters: " + ", ".join(method['deprecated_parameters']))


def main():
    cli_parser = argparse.ArgumentParser(
        description='Get Telegram API methods information from telegram '
                    'website.\n'
                    'Implement missing (or --all) methods in --out file, '
                    'or print methods information.',
        allow_abbrev=False,
    )
    cli_parser.add_argument('--file', '-f', '--filename', type=str,
                            default=None,
                            required=False,
                            help='File path to store Telegram API web page')
    cli_parser.add_argument('--all', '-a',
                            action='store_true',
                            help='Print all methods (default: print missing '
                                 'methods only)')
    cli_parser.add_argument('--out', '--output', '-o', type=str,
                            default=None,
                            required=False,
                            help='File path to store methods implementation')
    cli_parser.add_argument('--in', '--input', '-i', type=str,
                            default=None,
                            required=False,
                            help='File path to read Telegram API web page')
    cli_arguments = vars(cli_parser.parse_args())
    filename = cli_arguments['file']
    print_all = cli_arguments['all']
    output_file = cli_arguments['out']
    input_file = cli_arguments['in']
    asyncio.run(
        print_api_methods(filename=filename,
                          print_all=print_all,
                          output_file=output_file,
                          input_file=input_file)
    )
    logging.info("Done!")


if __name__ == '__main__':
    main()



"""Information about this package.

See `bot.py` for information about Bot class.
    ```python3.5+
    from davtelepot.bot import Bot
    help(Bot)
    ```
"""

__author__ = "Davide Testa"
__email__ = "davide@davte.it"
__credits__ = ["Marco Origlia", "Nick Lee @Nickoala"]
__license__ = "GNU General Public License v3.0"
__version__ = "2.8.11"
__maintainer__ = "Davide Testa"
__contact__ = "t.me/davte"

from davtelepot import (administration_tools, api, authorization,
                        bot, helper, languages, messages, suggestions,
                        useful_tools, utilities)

__all__ = ['administration_tools', 'api', 'authorization', 'bot', 'helper',
           'languages', 'messages', 'suggestions', 'useful_tools', 'utilities']



"""Use this tool to merge davtelepot-split files.

Documents larger than 50 MB are automatically split by `send_document` method
    and can be merged with this tool.
Example:
python3 -m davtelepot.tools.merge_files my_file.pdf
"""

import argparse
import glob
import logging
import os


def merge_files(input_file_path, output_file_path):
    input_directory = os.path.dirname(os.path.abspath(input_file_path))
    input_file_name = os.path.basename(os.path.abspath(input_file_path))
    if output_file_path:
        output_directory = os.path.dirname(os.path.abspath(output_file_path))
        output_file_name = os.path.basename(os.path.abspath(output_file_path))
    else:
        output_directory = input_directory
        output_file_name = input_file_name + 'out'
    with open(os.path.join(output_directory, output_file_name), 'wb') as output_file:
        for file_name in sorted(glob.glob(os.path.join(input_directory, input_file_name) + '*')):
            with open(file_name, 'rb') as input_file:
                output_file.write(input_file.read())


def main():
    # noinspection SpellCheckingInspection
    log_formatter = logging.Formatter(
        "%(asctime)s [%(module)-15s %(levelname)-8s]     %(message)s",
        style='%'
    )
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_formatter)
    console_handler.setLevel(logging.DEBUG)
    root_logger.addHandler(console_handler)

    # Parse command-line arguments
    cli_parser = argparse.ArgumentParser(description='Merge split files',
                                         allow_abbrev=False)
    cli_parser.add_argument('--input', '--file-path', '--path',
                            '-i', '-f', '-p',
                            type=str,
                            required=True,
                            help='Input file name (except ` - part n`)')
    cli_parser.add_argument('--output', '--output-file-path', '--out',
                            '-o',
                            type=str,
                            default=None,
                            required=False,
                            help='Output file name (defaults to input file name)')
    cli_arguments = vars(cli_parser.parse_args())
    input_file_path = cli_arguments['input']
    output_file_path = cli_arguments['output']
    merge_files(input_file_path=input_file_path,
                output_file_path=output_file_path)


if __name__ == '__main__':
    main()



"""Provide authorization levels to bot functions."""

# Standard library modules
from collections import OrderedDict
from typing import Callable, List, Union

# Project modules
from davtelepot.bot import Bot
from davtelepot.messages import default_authorization_messages
from davtelepot.utilities import (
    Confirmator, get_cleaned_text, get_user, make_button, make_inline_keyboard
)

DEFAULT_ROLES = OrderedDict()
DEFAULT_ROLES[0] = {
    'name': 'banned',
    'symbol': '🚫',
    'singular': 'banned',
    'plural': 'banned',
    'can_appoint': [],
    'can_be_appointed_by': [1, 2, 3]
}
DEFAULT_ROLES[1] = {
    'name': 'founder',
    'symbol': '👑',
    'singular': 'founder',
    'plural': 'founders',
    'can_appoint': [0, 1, 2, 3, 4, 5, 7, 100],
    'can_be_appointed_by': []
}
DEFAULT_ROLES[2] = {
    'name': 'admin',
    'symbol': '⚜️',
    'singular': 'administrator',
    'plural': 'administrators',
    'can_appoint': [0, 3, 4, 5, 7, 100],
    'can_be_appointed_by': [1]
}
DEFAULT_ROLES[3] = {
    'name': 'moderator',
    'symbol': '🔰',
    'singular': 'moderator',
    'plural': 'moderators',
    'can_appoint': [0, 5, 7],
    'can_be_appointed_by': [1, 2]
}
DEFAULT_ROLES[5] = {
    'name': 'user',
    'symbol': '🎫',
    'singular': 'registered user',
    'plural': 'registered users',
    'can_appoint': [],
    'can_be_appointed_by': [1, 2, 3]
}
DEFAULT_ROLES[100] = {
    'name': 'everybody',
    'symbol': '👤',
    'singular': 'common user',
    'plural': 'common users',
    'can_appoint': [],
    'can_be_appointed_by': [1, 2, 3]
}

max_records_per_query = 15


class Role:
    """Authorization level for users of a bot."""

    roles = OrderedDict()
    default_role_code = 100

    def __init__(self, code: int, name: str, symbol: str,
                 singular: str, plural: str,
                 can_appoint: List[int], can_be_appointed_by: List[int]):
        """Instantiate Role object.

        code : int
            The higher the code, the fewer privileges are connected to that
                role. Use 0 for banned users.
        name : str
            Short name for role.
        symbol : str
            Emoji used to represent role.
        singular : str
            Singular full name of role.
        plural : str
            Plural full name of role.
        can_appoint : list of int
            List of role codes that this role can appoint.
        can_be_appointed_by : list of int
            List of role codes this role can be appointed by.
        """
        self._code = code
        self._name = name
        self._symbol = symbol
        self._singular = singular
        self._plural = plural
        self._can_appoint = can_appoint
        self._can_be_appointed_by = can_be_appointed_by
        self.__class__.roles[self.code] = self

    @property
    def code(self) -> int:
        """Return code."""
        return self._code

    @property
    def name(self) -> str:
        """Return name."""
        return self._name

    @property
    def symbol(self) -> str:
        """Return symbol."""
        return self._symbol

    @property
    def singular(self) -> str:
        """Return singular."""
        return self._singular

    @property
    def plural(self) -> str:
        """Return plural."""
        return self._plural

    @property
    def can_appoint(self) -> List[int]:
        """Return can_appoint."""
        return self._can_appoint

    @property
    def can_be_appointed_by(self) -> List[int]:
        """Return roles whom this role can be appointed by."""
        return self._can_be_appointed_by

    @classmethod
    def get_by_role_id(cls, role_id=100) -> 'Role':
        """Given a `role_id`, return the corresponding `Role` instance."""
        for code, role in cls.roles.items():
            if code == role_id:
                return role
        raise IndexError(f"Unknown role id: {role_id}")

    @classmethod
    def get_role_by_name(cls, name='everybody') -> 'Role':
        """Given a `name`, return the corresponding `Role` instance."""
        for role in cls.roles.values():
            if role.name == name:
                return role
        raise IndexError(f"Unknown role name: {name}")

    @classmethod
    def get_user_role(cls,
                      user_record: OrderedDict = None,
                      user_role_id: int = None) -> 'Role':
        """Given a `user_record`, return its `Role`.

        `role_id` may be passed as keyword argument or as user_record.
        """
        if user_role_id is None:
            if isinstance(user_record, dict) and 'privileges' in user_record:
                user_role_id = user_record['privileges']
            elif type(user_record) is int:
                user_role_id = user_record
        if type(user_role_id) is not int:
            for code, role in cls.roles.items():
                if role.name == user_role_id:
                    user_role_id = code
                    break
            else:
                user_role_id = cls.default_role_code
        return cls.get_by_role_id(role_id=user_role_id)

    @classmethod
    def set_default_role_code(cls, role: int) -> None:
        """Set class default role code.

        It will be returned if a specific role code cannot be evaluated.
        """
        cls.default_role_code = role

    @classmethod
    def get_user_role_text(cls,
                           user_record: OrderedDict,
                           user_role: 'Role' = None) -> str:
        """
        Get a string to describe the role of a user.

        @param user_record: record of table `users` about the user; it must
            contain at least a [username | last_name | first_name] and a
            telegram identifier.
        @param user_role: Role instance about user permissions.
        @return: String to describe the role of a user, like this:
            ```
            👤 LinkedUsername
            🔑 Admin ⚜️
            ```
        """
        if user_role is None:
            user_role = cls.get_user_role(user_record=user_record)
        long_name = ' '.join(
            [user_record[key]
             for key in ('first_name', 'last_name')
             if key in user_record
             and user_record[key]]
        )
        if long_name:
            long_name = f" ({long_name})"
        return (
            f"👤 {get_user(record=user_record)}{long_name}\n"
            f"🔑 <i>{user_role.singular.capitalize()}</i> {user_role.symbol}"
        )

    @classmethod
    def get_user_role_buttons(cls,
                              user_record: OrderedDict,
                              admin_record: OrderedDict,
                              user_role: 'Role' = None,
                              admin_role: 'Role' = None) -> List[dict]:
        """ Return buttons to edit user permissions.
        @param user_record: record of table `users` about the user; it must
            contain at least a [username | last_name | first_name] and a
            telegram identifier.
        @param admin_record: record of table `users` about the admin; it must
            contain at least a [username | last_name | first_name] and a
            telegram identifier.
        @param user_role: Role instance about user permissions.
        @param admin_role: Role instance about admin permissions.
        @return: list of `InlineKeyboardButton`s.
        """
        if admin_role is None:
            admin_role = cls.get_user_role(user_record=admin_record)
        if user_role is None:
            user_role = cls.get_user_role(user_record=user_record)
        return [
            make_button(
                f"{role.symbol} {role.singular.capitalize()}",
                prefix='auth:///',
                delimiter='|',
                data=['set', user_record['id'], code]
            )
            for code, role in cls.roles.items()
            if (admin_role > user_role
                and code in admin_role.can_appoint
                and code != user_role.code)
        ]

    @classmethod
    def get_user_role_text_and_buttons(cls,
                                       user_record: OrderedDict,
                                       admin_record: OrderedDict):
        """Get text and buttons for user role panel."""
        admin_role = cls.get_user_role(user_record=admin_record)
        user_role = cls.get_user_role(user_record=user_record)
        text = cls.get_user_role_text(user_record=user_record,
                                      user_role=user_role)
        buttons = cls.get_user_role_buttons(user_record=user_record,
                                            user_role=user_role,
                                            admin_record=admin_record,
                                            admin_role=admin_role)
        return text, buttons

    def __eq__(self, other: 'Role'):
        """Return True if `self` is equal to `other`."""
        return self.code == other.code

    def __gt__(self, other: 'Role'):
        """Return True if self can appoint other."""
        return (
                (
                        self.code < other.code
                        or other.code == 0
                )
                and self.code in other.can_be_appointed_by
        )

    def __ge__(self, other: 'Role'):
        """Return True if self >= other."""
        return self.__gt__(other) or self.__eq__(other)

    def __lt__(self, other: 'Role'):
        """Return True if self can not appoint other."""
        return not self.__ge__(other)

    def __le__(self, other: 'Role'):
        """Return True if `self` is superior or equal to `other`."""
        return not self.__gt__(other)

    def __ne__(self, other: 'Role'):
        """Return True if `self` is not equal to `other`."""
        return not self.__eq__(other)

    def __str__(self):
        """Return human-readable description of role."""
        return f"<Role object: {self.symbol} {self.singular.capitalize()}>"


def get_authorization_function(bot: Bot):
    """Take a `bot` and return its authorization_function."""

    def is_authorized(update, user_record=None, authorization_level=2):
        """Return True if user role is at least at `authorization_level`."""
        if user_record is None:
            if (
                    isinstance(update, dict)
                    and 'from' in update
                    and isinstance(update['from'], dict)
                    and 'id' in update['from']
            ):
                user_record = bot.db['users'].find_one(
                    telegram_id=update['from']['id']
                )
        user_role = bot.Role.get_user_role(user_record=user_record)
        if user_role.code == 0:
            return False
        needed_role = bot.Role.get_user_role(user_role_id=authorization_level)
        if needed_role.code < user_role.code:
            return False
        return True

    return is_authorized


def get_browse_buttons(bot: Bot, language: str):
    return [
        make_button(
            text=bot.get_message('authorization', 'auth_button',
                                 'browse', 'browse_button_az',
                                 language=language),
            prefix='auth:///',
            delimiter='|',
            data=['browse', 'az'],
        ),
        make_button(
            text=bot.get_message('authorization', 'auth_button',
                                 'browse', 'browse_button_by_role',
                                 language=language),
            prefix='auth:///',
            delimiter='|',
            data=['browse', 'role'],
        ),
    ]


async def _authorization_command(bot: Bot,
                                 update: dict,
                                 user_record: OrderedDict,
                                 language: str,
                                 mode: str = 'auth'):
    db = bot.db
    text = get_cleaned_text(bot=bot, update=update, replace=[mode])
    reply_markup = None
    admin_record = user_record.copy()
    query_id = None
    user_record = None
    admin_role = bot.Role.get_user_role(user_record=admin_record)
    result = bot.get_message(
        'authorization', 'auth_command', 'unhandled_case',
        update=update, user_record=admin_record
    )
    if not text:  # No text provided: command must be used in reply
        if 'reply_to_message' not in update:  # No text and not in reply
            result = bot.get_message(
                'authorization', 'auth_command', 'instructions',
                update=update, user_record=admin_record,
                command=mode
            )
            buttons = get_browse_buttons(bot=bot, language=language)
            reply_markup = make_inline_keyboard(buttons, 2)
            user_record = -1
        else:  # No text, command used in reply to another message
            update = update['reply_to_message']
            # Forwarded message: get both the user who forwarded and the original author
            if ('forward_from' in update
                    and update['from']['id'] != update['forward_from']['id']):
                user_record = list(
                    db['users'].find(
                        telegram_id=[update['from']['id'],
                                     update['forward_from']['id']]
                    )
                )
            else:  # Otherwise: get the author of the message
                user_record = db['users'].find_one(
                    telegram_id=update['from']['id']
                )
    else:  # Get users matching the input text
        query_text = (
            "SELECT * "
            "FROM users "
            "WHERE COALESCE("
            "   first_name || last_name || username,"
            "   last_name || username,"
            "   first_name || username,"
            "   username,"
            "   first_name || last_name,"
            "   last_name,"
            "   first_name"
            f") LIKE '%{text}%' "
            "ORDER BY COALESCE("
            "   username || first_name || last_name,"
            "   username || last_name,"
            "   username || first_name,"
            "   username,"
            "   first_name || last_name,"
            "   first_name,"
            "   last_name"
            ")"
        )
        query_id = bot.db['queries'].upsert(
            dict(query=query_text),
            ['query']
        )
        if query_id is True:
            query_id = bot.db['queries'].find_one(
                query=query_text
            )['id']
        user_record = list(
            db.query(
                "SELECT * "
                "FROM users "
                "WHERE COALESCE("
                "   first_name || last_name || username,"
                "   last_name || username,"
                "   first_name || username,"
                "   username,"
                "   first_name || last_name,"
                "   last_name,"
                "   first_name"
                f") LIKE '%{text}%'"
            )
        )
        if len(user_record) == 1:
            user_record = user_record[0]
    if user_record is None:  # If query was not provided and user cannot be found
        result = bot.get_message(
            'authorization', 'auth_command', 'unknown_user',
            update=update, user_record=admin_record
        )
    elif type(user_record) is list and len(user_record) > 1:  # If many users match
        browse_panel = await _authorization_button(bot=bot,
                                                   update=update,
                                                   user_record=admin_record,
                                                   language=language,
                                                   data=['browse', query_id])
        if 'edit' in browse_panel:
            return browse_panel['edit']
    elif type(user_record) is list and len(user_record) == 0:  # If query was provided but no user matches
        result = bot.get_message(
            'authorization', 'auth_command', 'no_match',
            update=update, user_record=admin_record,
        )
    elif (type(user_record) is list and len(user_record) == 1) or isinstance(user_record, dict):  # If 1 user matches
        if type(user_record) is list:
            user_record = user_record[0]
        # Ban user if admin can do it
        user_role = bot.Role.get_user_role(user_record=user_record)
        if mode == 'ban' and admin_role > user_role:
            user_record['privileges'] = 0
            db['users'].update(
                user_record,
                ['id']
            )
        # Show user panel (text and buttons) to edit user permissions
        result, buttons = bot.Role.get_user_role_text_and_buttons(
            user_record=user_record,
            admin_record=admin_record
        )
        if bot.db['user_profile_photos'].find_one(user_id=user_record['id']):
            buttons.append(
                make_button(
                    text=bot.get_message('authorization', 'auth_button',
                                         'profile_picture_button',
                                         language=language),
                    prefix='auth:///',
                    delimiter='|',
                    data=['picture', user_record['id']]
                )
            )
        reply_markup = make_inline_keyboard(buttons, 1)
        reply_markup['inline_keyboard'].append(get_browse_buttons(bot=bot, language=language))
    return dict(
        text=result,
        reply_markup=reply_markup,
        parse_mode='HTML'
    )


async def _authorization_button(bot: Bot,
                                update: dict,
                                user_record: OrderedDict,
                                language: str,
                                data: Union[str, List[Union[int, str]]]):
    if len(data) == 0:
        data = ['']
    command, *arguments = data
    user_id = user_record['telegram_id']
    if len(arguments) > 0 and command in ['show', 'set']:
        other_user_id = arguments[0]
    else:
        other_user_id = None
    result, text, reply_markup = '', '', None
    query_text = None
    if command in ['browse'] and len(arguments) >= 1:
        mode = arguments[0]
        offset = arguments[1] if len(arguments) > 1 else 0
        user_records = []
        if mode == 'choose':
            update['text'] = ''
            answer_update = await _authorization_command(bot=bot,
                                                         update=update,
                                                         user_record=user_record,
                                                         language=language,
                                                         mode='auth')
            text = answer_update['text']
            reply_markup = answer_update['reply_markup']
        elif isinstance(mode, int) or (isinstance(mode, str) and mode.isnumeric()):
            query_record = bot.db['queries'].find_one(id=int(mode))
            if query_record:
                query_text = query_record['query']
        elif mode == 'az':
            query_text = (
                "SELECT * "
                "FROM users "
                "ORDER BY COALESCE("
                "   username || first_name || last_name,"
                "   username || last_name,"
                "   username || first_name,"
                "   username,"
                "   first_name || last_name,"
                "   first_name,"
                "   last_name"
                ")"
            )
        elif mode == 'role':
            query_text = (
                "SELECT * "
                "FROM users "
                "ORDER BY ("
                "   CASE WHEN privileges = 0 THEN 5000 "
                "   ELSE privileges END) "
                "ASC, COALESCE( "
                "   username || first_name || last_name,"
                "   username || last_name,"
                "   username || first_name,"
                "   username,"
                "   first_name || last_name,"
                "   first_name,"
                "   last_name"
                ") "
            )
        n = 0
        if query_text:
            user_records = list(
                bot.db.query(
                    f"{query_text} "
                    f"LIMIT {max_records_per_query + 1} "
                    f"OFFSET {offset * max_records_per_query} "
                )
            )
            for record in bot.db.query("SELECT COUNT(*) n "
                                       f"FROM ({query_text})"):
                n = record['n']
        if user_records:
            text = bot.get_message(
                'authorization', 'auth_command', 'choose_user',
                update=update, user_record=user_record,
                n=n
            )
            reply_markup = make_inline_keyboard(
                [
                    make_button(
                        f"{bot.Role.get_user_role(user_record=user).symbol} {get_user(user, link_profile=False)}",
                        prefix='auth:///',
                        delimiter='|',
                        data=['show', user['id'], command, mode, offset]
                    )
                    for user in user_records[:max_records_per_query]
                ],
                3
            )
            if n > max_records_per_query:
                reply_markup['inline_keyboard'].append(
                    [
                        make_button(
                            text='◀️',
                            prefix='auth:///',
                            delimiter='|',
                            data=['browse', mode, offset - 1 if offset else n // max_records_per_query]
                        ),
                        make_button(
                            text='↩️',
                            prefix='auth:///',
                            delimiter='|',
                            data=['browse', 'choose']
                        ),
                        make_button(
                            text='▶️',
                            prefix='auth:///',
                            delimiter='|',
                            data=['browse', mode, offset + 1 if n > (offset + 1) * max_records_per_query else 0]
                        )
                    ]
                )
    elif command in ['show']:
        other_user_record = bot.db['users'].find_one(id=other_user_id)
        text, buttons = bot.Role.get_user_role_text_and_buttons(
            user_record=other_user_record,
            admin_record=user_record
        )
        if bot.db['user_profile_photos'].find_one(user_id=other_user_record['id']):
            buttons.append(
                make_button(
                    text=bot.get_message('authorization', 'auth_button',
                                         'profile_picture_button',
                                         language=language),
                    prefix='auth:///',
                    delimiter='|',
                    data=['picture', other_user_record['id']]
                )
            )
        if len(arguments) > 2:
            buttons.append(
                make_button(
                    text='↩️',
                    prefix='auth:///',
                    delimiter='|',
                    data=data[2:]
                )
            )
        reply_markup = make_inline_keyboard(buttons, 1)
    elif command in ['set'] and len(arguments) > 1:
        other_user_id, new_privileges, *_ = arguments
        if not Confirmator.get(
                key=f'{user_id}_set_{other_user_id}',
                confirm_timedelta=5
        ).confirm:
            return bot.get_message(
                'authorization', 'auth_button', 'confirm',
                update=update, user_record=user_record,
            )
        other_user_record = bot.db['users'].find_one(id=other_user_id)
        user_role = bot.Role.get_user_role(user_record=user_record)
        other_user_role = bot.Role.get_user_role(user_record=other_user_record)
        if other_user_role.code == new_privileges:
            return bot.get_message(
                'authorization', 'auth_button', 'no_change',
                update=update, user_record=user_record
            )
        if not user_role > other_user_role:
            text = bot.get_message(
                'authorization', 'auth_button', 'permission_denied', 'user',
                update=update, user_record=user_record
            )
            reply_markup = make_inline_keyboard(
                [
                    make_button(
                        bot.get_message(
                            'authorization', 'auth_button', 'back_to_user',
                            update=update, user_record=user_record
                        ),
                        prefix='auth:///',
                        delimiter='|',
                        data=['show', other_user_id]
                    )
                ],
                1
            )
        elif new_privileges not in user_role.can_appoint:
            text = bot.get_message(
                'authorization', 'auth_button', 'permission_denied', 'role',
                update=update, user_record=user_record
            )
            reply_markup = make_inline_keyboard(
                [
                    make_button(
                        bot.get_message(
                            'authorization', 'auth_button', 'back_to_user',
                            update=update, user_record=user_record
                        ),
                        prefix='auth:///',
                        delimiter='|',
                        data=['show', other_user_id]
                    )
                ],
                1
            )
        else:
            bot.db['users'].update(
                dict(
                    id=other_user_id,
                    privileges=new_privileges
                ),
                ['id']
            )
            other_user_record = bot.db['users'].find_one(id=other_user_id)
            result = bot.get_message(
                'authorization', 'auth_button', 'appointed',
                update=update, user_record=user_record
            )
            text, buttons = bot.Role.get_user_role_text_and_buttons(
                user_record=other_user_record,
                admin_record=user_record
            )
            if bot.db['user_profile_photos'].find_one(user_id=other_user_record['id']):
                buttons.append(
                    make_button(
                        text=bot.get_message('authorization', 'auth_button',
                                             'profile_picture_button',
                                             language=language),
                        prefix='auth:///',
                        delimiter='|',
                        data=['picture', other_user_record['id']]
                    )
                )
            reply_markup = make_inline_keyboard(buttons, 1)
    elif command in ['picture'] and len(arguments) > 0:
        photo_record = bot.db['user_profile_photos'].find_one(
            user_id=arguments[0],
            order_by=['-update_datetime'],
        )
        other_user_record = bot.db['users'].find_one(id=arguments[0])
        if photo_record is None:
            result = bot.get_message('admin', 'error', 'text',
                                     language=language)
        else:
            caption, buttons = bot.Role.get_user_role_text_and_buttons(
                user_record=other_user_record,
                admin_record=user_record
            )
            await bot.sendPhoto(
                chat_id=user_record['telegram_id'],
                photo=photo_record['telegram_file_id'],
                caption=caption,
                reply_markup=make_inline_keyboard(
                    buttons=buttons
                ),
                parse_mode='HTML'
            )
    if text:
        return dict(
            text=result,
            edit=dict(
                text=text,
                reply_markup=reply_markup,
                parse_mode='HTML'
            )
        )
    return result


def default_get_administrators_function(bot: Bot):
    return list(
        bot.db['users'].find(privileges=[1, 2])
    )


def init(telegram_bot: Bot,
         roles: Union[list, OrderedDict] = None,
         authorization_messages=None,
         get_administrators_function: Callable[[object],
                                               list] = None):
    """Set bot roles and assign role-related commands.

    Pass an OrderedDict of `roles` to get them set.
    """

    class _Role(Role):
        roles = OrderedDict()

    telegram_bot.set_role_class(_Role)
    if roles is None:
        roles = DEFAULT_ROLES
    # Cast roles to OrderedDict
    if isinstance(roles, list):
        roles = OrderedDict(
            (i, element)
            for i, element in enumerate(roles)
        )
    if not isinstance(roles, OrderedDict):
        raise TypeError("`roles` shall be a OrderedDict!")
    for code, role in roles.items():
        if 'code' not in role:
            role['code'] = code
        telegram_bot.Role(**role)

    telegram_bot.set_authorization_function(
        get_authorization_function(telegram_bot)
    )
    get_administrators_function = (get_administrators_function
                                   or default_get_administrators_function)
    telegram_bot.set_get_administrator_function(get_administrators_function)
    authorization_messages = (authorization_messages
                              or default_authorization_messages)
    telegram_bot.messages['authorization'] = authorization_messages

    @telegram_bot.command(command='/auth', aliases=[], show_in_keyboard=False,
                          description=(
                                  authorization_messages['auth_command']['description']
                          ),
                          authorization_level='moderator')
    async def authorization_command(bot, update, user_record, language):
        return await _authorization_command(bot=bot, update=update,
                                            user_record=user_record,
                                            language=language)

    @telegram_bot.button('auth:///',
                         description=authorization_messages['auth_button']['description'],
                         separator='|',
                         authorization_level='moderator')
    async def authorization_button(bot, update, user_record, language, data):
        return await _authorization_button(bot=bot, update=update,
                                           user_record=user_record,
                                           language=language, data=data)

    @telegram_bot.command('/ban', aliases=[], show_in_keyboard=False,
                          description=authorization_messages['ban_command']['description'],
                          authorization_level='moderator')
    async def ban_command(bot, update, user_record, language):
        return await _authorization_command(bot=bot, update=update,
                                            user_record=user_record,
                                            language=language, mode='ban')
